//! NUT-02: Keysets and keyset ID
//!
//! <https://github.com/cashubtc/nuts/blob/main/02.md>

use core::fmt;
use core::str::FromStr;
use std::array::TryFromSliceError;
use std::collections::BTreeMap;

use bitcoin::bip32::DerivationPath;
use bitcoin::bip32::{ChildNumber, Xpriv};
use bitcoin::hashes::sha256::Hash as Sha256;
use bitcoin::hashes::Hash;
use bitcoin::key::Secp256k1;
use bitcoin::secp256k1;
use serde::{Deserialize, Serialize};
use serde_with::{serde_as, VecSkipError};
use thiserror::Error;

use super::nut01::Keys;
use super::nut01::{MintKeyPair, MintKeys};
use crate::amount::AmountStr;
use crate::hex;
use crate::nuts::nut00::CurrencyUnit;
use crate::Amount;

/// NUT02 Error
#[derive(Debug, Error)]
pub enum Error {
    /// Hex Error
    #[error(transparent)]
    HexError(#[from] hex::Error),
    /// Keyset length error
    #[error("NUT02: ID length invalid")]
    Length,
    /// Unknown version
    #[error("NUT02: Unknown Version")]
    UnknownVersion,
    /// Keyset id does not match
    #[error("Keyset id incorrect")]
    IncorrectKeysetId,
    /// Slice Error
    #[error(transparent)]
    Slice(#[from] TryFromSliceError),
}

/// Keyset version
#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
#[cfg_attr(feature = "swagger", derive(utoipa::ToSchema))]
pub enum KeySetVersion {
    /// Current Version 00
    Version00,
}

impl KeySetVersion {
    /// [`KeySetVersion`] to byte
    pub fn to_byte(&self) -> u8 {
        match self {
            Self::Version00 => 0,
        }
    }

    /// [`KeySetVersion`] from byte
    pub fn from_byte(byte: &u8) -> Result<Self, Error> {
        match byte {
            0 => Ok(Self::Version00),
            _ => Err(Error::UnknownVersion),
        }
    }
}

impl fmt::Display for KeySetVersion {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            KeySetVersion::Version00 => f.write_str("00"),
        }
    }
}

/// A keyset ID is an identifier for a specific keyset. It can be derived by
/// anyone who knows the set of public keys of a mint. The keyset ID **CAN**
/// be stored in a Cashu token such that the token can be used to identify
/// which mint or keyset it was generated from.
#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash, Serialize, Deserialize)]
#[serde(into = "String", try_from = "String")]
#[cfg_attr(feature = "swagger", derive(utoipa::ToSchema), schema(as = String))]
pub struct Id {
    version: KeySetVersion,
    id: [u8; Self::BYTELEN],
}

impl Id {
    const STRLEN: usize = 14;
    const BYTELEN: usize = 7;

    /// [`Id`] to bytes
    pub fn to_bytes(&self) -> Vec<u8> {
        [vec![self.version.to_byte()], self.id.to_vec()].concat()
    }

    /// [`Id`] from bytes
    pub fn from_bytes(bytes: &[u8]) -> Result<Self, Error> {
        Ok(Self {
            version: KeySetVersion::from_byte(&bytes[0])?,
            id: bytes[1..].try_into()?,
        })
    }

    /// [`Id`] as bytes
    pub fn as_bytes(&self) -> [u8; Self::BYTELEN + 1] {
        let mut bytes = [0u8; Self::BYTELEN + 1];
        bytes[0] = self.version.to_byte();
        bytes[1..].copy_from_slice(&self.id);
        bytes
    }
}

// Used to generate a compressed unique identifier as part of the NUT13 spec
// This is a one-way function
impl From<Id> for u32 {
    fn from(value: Id) -> Self {
        let hex_bytes: [u8; 8] = value.as_bytes();

        let int = u64::from_be_bytes(hex_bytes);

        (int % (2_u64.pow(31) - 1)) as u32
    }
}

impl fmt::Display for Id {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.write_str(&format!("{}{}", self.version, hex::encode(self.id)))
    }
}

impl TryFrom<String> for Id {
    type Error = Error;

    fn try_from(s: String) -> Result<Self, Self::Error> {
        if s.len() != 16 {
            return Err(Error::Length);
        }

        Ok(Self {
            version: KeySetVersion::from_byte(&hex::decode(&s[..2])?[0])?,
            id: hex::decode(&s[2..])?
                .try_into()
                .map_err(|_| Error::Length)?,
        })
    }
}

impl FromStr for Id {
    type Err = Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Self::try_from(s.to_string())
    }
}

impl From<Id> for String {
    fn from(value: Id) -> Self {
        value.to_string()
    }
}

impl From<&Keys> for Id {
    /// As per NUT-02:
    ///   1. sort public keys by their amount in ascending order
    ///   2. concatenate all public keys to one string
    ///   3. HASH_SHA256 the concatenated public keys
    ///   4. take the first 14 characters of the hex-encoded hash
    ///   5. prefix it with a keyset ID version byte
    fn from(map: &Keys) -> Self {
        let mut keys: Vec<(&AmountStr, &super::PublicKey)> = map.iter().collect();
        keys.sort_by_key(|(amt, _v)| *amt);

        let pubkeys_concat: Vec<u8> = keys
            .iter()
            .map(|(_, pubkey)| pubkey.to_bytes())
            .collect::<Vec<[u8; 33]>>()
            .concat();

        let hash = Sha256::hash(&pubkeys_concat);
        let hex_of_hash = hex::encode(hash.to_byte_array());

        Self {
            version: KeySetVersion::Version00,
            id: hex::decode(&hex_of_hash[0..Self::STRLEN])
                .expect("Keys hash could not be hex decoded")
                .try_into()
                .expect("Invalid length of hex id"),
        }
    }
}

/// Mint Keysets [NUT-02]
/// Ids of mints keyset ids
#[serde_as]
#[derive(Debug, Clone, PartialEq, Eq, Serialize, Deserialize)]
#[cfg_attr(feature = "swagger", derive(utoipa::ToSchema))]
pub struct KeysetResponse {
    /// set of public key ids that the mint generates
    #[serde_as(as = "VecSkipError<_>")]
    pub keysets: Vec<KeySetInfo>,
}

/// Keyset
#[derive(Debug, Clone, PartialEq, Eq, Deserialize, Serialize)]
#[cfg_attr(feature = "swagger", derive(utoipa::ToSchema))]
pub struct KeySet {
    /// Keyset [`Id`]
    pub id: Id,
    /// Keyset [`CurrencyUnit`]
    pub unit: CurrencyUnit,
    /// Keyset [`Keys`]
    pub keys: Keys,
}

impl KeySet {
    /// Verify the keyset is matches keys
    pub fn verify_id(&self) -> Result<(), Error> {
        let keys_id: Id = (&self.keys).into();

        if keys_id != self.id {
            return Err(Error::IncorrectKeysetId);
        }

        Ok(())
    }
}

impl From<MintKeySet> for KeySet {
    fn from(keyset: MintKeySet) -> Self {
        Self {
            id: keyset.id,
            unit: keyset.unit,
            keys: Keys::from(keyset.keys),
        }
    }
}

/// KeySetInfo
#[derive(Debug, Clone, Hash, PartialEq, Eq, Deserialize, Serialize)]
#[cfg_attr(feature = "swagger", derive(utoipa::ToSchema))]
pub struct KeySetInfo {
    /// Keyset [`Id`]
    pub id: Id,
    /// Keyset [`CurrencyUnit`]
    pub unit: CurrencyUnit,
    /// Keyset state
    /// Mint will only sign from an active keyset
    pub active: bool,
    /// Input Fee PPK
    #[serde(default = "default_input_fee_ppk")]
    pub input_fee_ppk: u64,
}

fn default_input_fee_ppk() -> u64 {
    0
}

/// MintKeyset
#[derive(Debug, Clone, PartialEq, Eq, Serialize, Deserialize)]
pub struct MintKeySet {
    /// Keyset [`Id`]
    pub id: Id,
    /// Keyset [`CurrencyUnit`]
    pub unit: CurrencyUnit,
    /// Keyset [`MintKeys`]
    pub keys: MintKeys,
}

impl From<MintKeySet> for Id {
    fn from(keyset: MintKeySet) -> Id {
        let keys: super::KeySet = keyset.into();

        Id::from(&keys.keys)
    }
}

impl From<&MintKeys> for Id {
    fn from(map: &MintKeys) -> Self {
        let keys: super::Keys = map.clone().into();

        Id::from(&keys)
    }
}
