//! NUT-12: Offline ecash signature validation
//!
//! <https://github.com/cashubtc/nuts/blob/main/12.md>

use bitcoin::secp256k1;
use serde::{Deserialize, Serialize};
use thiserror::Error;

use super::nut01::SecretKey;

/// NUT12 Error
#[derive(Debug, Error)]
pub enum Error {
    /// Missing DLEQ Proof
    #[error("No DLEQ proof provided")]
    MissingDleqProof,
    /// Incomplete DLEQ Proof
    #[error("Incomplete DLEQ proof")]
    IncompleteDleqProof,
    /// Invalid DLEQ Proof
    #[error("Invalid DLEQ proof")]
    InvalidDleqProof,
    /// DHKE error
    #[error(transparent)]
    DHKE(#[from] crate::dhke::Error),
    /// NUT01 Error
    #[error(transparent)]
    NUT01(#[from] crate::nuts::nut01::Error),
    /// SECP256k1 Error
    #[error(transparent)]
    Secp256k1(#[from] secp256k1::Error),
}

/// Blinded Signature on Dleq
///
/// Defined in [NUT12](https://github.com/cashubtc/nuts/blob/main/12.md)
#[derive(Debug, Clone, PartialEq, Eq, Serialize, Deserialize)]
#[cfg_attr(feature = "swagger", derive(utoipa::ToSchema))]
pub struct BlindSignatureDleq {
    /// e
    #[cfg_attr(feature = "swagger", schema(value_type = String))]
    pub e: SecretKey,
    /// s
    #[cfg_attr(feature = "swagger", schema(value_type = String))]
    pub s: SecretKey,
}

/// Proof Dleq
///
/// Defined in [NUT12](https://github.com/cashubtc/nuts/blob/main/12.md)
#[derive(Debug, Clone, PartialEq, Eq, Serialize, Deserialize)]
#[cfg_attr(feature = "swagger", derive(utoipa::ToSchema))]
pub struct ProofDleq {
    /// e
    #[cfg_attr(feature = "swagger", schema(value_type = String))]
    pub e: SecretKey,
    /// s
    #[cfg_attr(feature = "swagger", schema(value_type = String))]
    pub s: SecretKey,
    /// Blinding factor
    #[cfg_attr(feature = "swagger", schema(value_type = String))]
    pub r: SecretKey,
}
